﻿using System;
using System.IO;
using System.Reflection;
using System.Runtime.InteropServices.WindowsRuntime;
using TNKLabo.S1D13781;
using Windows.Foundation;
using Windows.Graphics.Imaging;
using Windows.Storage.Streams;
using Windows.UI;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Imaging;
using Windows.UI.Xaml.Navigation;
using Windows.UI.Xaml.Shapes;

namespace EpsonLcdDemo
{
	public sealed partial class MainPage : Page
	{
		private enum DrawMode
		{
			Pixel, Line, Rect, Text, Bitmap
		}

		private S1D13781 s1d13781;
		private DrawMode drawmode = DrawMode.Pixel;
		private ColorPicker.ColorPicker cp = null;
		private Color currentColor = Colors.White;
		private byte[] sampleImage = null;
		private int siWidth, siHeight;

		public MainPage()
		{
			this.InitializeComponent();

			cmbType.Items.Add("Pixel");
			cmbType.Items.Add("Line");
			cmbType.Items.Add("FillRect");
			cmbType.Items.Add("Text");
			cmbType.Items.Add("Bitmap");
			cmbType.SelectedIndex = 0;

			UpdateButtonColor();
		}

		private void OnSelectionChanged(object sender, SelectionChangedEventArgs e)
		{
			txtSample.IsEnabled = (cmbType.SelectedIndex == 3);
			switch (cmbType.SelectedIndex) {
			case 1: drawmode = DrawMode.Line; break;
			case 2: drawmode = DrawMode.Rect; break;
			case 3: drawmode = DrawMode.Text; break;
			case 4:
				drawmode = DrawMode.Bitmap;
				if (sampleImage == null) LoadBitmap();
				break;
			default: drawmode = DrawMode.Pixel; break;
			}
		}

		private async void LoadBitmap()
		{
			IRandomAccessStream rs = await RandomAccessStreamReference.CreateFromUri
				(new Uri("ms-appx:///Assets/sample.jpg")).OpenReadAsync();

			BitmapDecoder decoder = await BitmapDecoder.CreateAsync(rs);
			PixelDataProvider pixelData = await decoder.GetPixelDataAsync();
			siWidth = (int)decoder.PixelWidth;
			siHeight = (int)decoder.PixelHeight;
			sampleImage = pixelData.DetachPixelData();
		}

		protected override void OnNavigatedTo(NavigationEventArgs e)
		{
			base.OnNavigatedTo(e);

			s1d13781 = new S1D13781();
			s1d13781.Begin();

			if(s1d13781.Width == 0) {
				// For UI debugging
				canvas.Width = 482;
				canvas.Height = 274;
			} else {
				canvas.Width = s1d13781.Width + 2;
				canvas.Height = s1d13781.Height + 2;
			}
			drawarea.Width = canvas.Width - 2;
			drawarea.Height = canvas.Height - 2;
		}

		private void OnClickClear(object sender, RoutedEventArgs e)
		{
			s1d13781.Clear();
		}

		private void UpdateButtonColor()
		{
			SolidColorBrush bgb = new SolidColorBrush(currentColor);
			bgb.Opacity = 128;
			btnColor.Background = bgb;
		}

		private void OnClickColor(object sender, RoutedEventArgs e)
		{
			if (cp != null) return;

			cp = new ColorPicker.ColorPicker();
			cp.Height = 480;
			cp.Width = 640;
			cp.HorizontalAlignment = HorizontalAlignment.Center;
			cp.VerticalAlignment = VerticalAlignment.Center;
			cp.DoubleTapped += new DoubleTappedEventHandler((o, d) => {
				currentColor = cp.SelectedColor.Color;
				UpdateButtonColor();

				maingrid.Children.Remove(cp);
				cp = null;
			});
			maingrid.Children.Add(cp);
		}

		private struct uPoint
		{
			public ushort x, y;
			public uPoint(ushort x, ushort y)
			{
				this.x = x;
				this.y = y;
			}
			public uPoint(Point pt)
			{
				if (pt.X < 0) x = 0; else x = (ushort)(pt.X + 0.5);
				if (pt.Y < 0) y = 0; else y = (ushort)(pt.Y + 0.5);
			}
		}

		private bool pressed = false;
		private const double ecr = 5.0;
		private Line line_guide = null;
		private Rectangle rect_guide = null;
		private TextBlock text_guide = null;
		private uPoint pt1;

		private void OnDrawAreaPressed(object sender, PointerRoutedEventArgs e)
		{
			ClearGuides();

			DispatcherTimer tmr;

			pt1 = new uPoint(e.GetCurrentPoint(drawarea).Position);
			switch (drawmode) {
			case DrawMode.Pixel:
				s1d13781.DrawPixel(pt1.x, pt1.y, currentColor);

				var ec = new Ellipse();
				ec.Margin = new Thickness(pt1.x - ecr / 2.0, pt1.y - ecr / 2.0, 0, 0);
				ec.Fill = new SolidColorBrush(currentColor);
				ec.Width = ecr;
				ec.Height = ecr;
				drawarea.Children.Add(ec);
				tmr = new DispatcherTimer();
				tmr.Interval = new TimeSpan(0,0,0,750);
				tmr.Tick += (s, o) => {
					tmr.Stop();
					drawarea.Children.Remove(ec);
				};
				tmr.Start();
				break;
			case DrawMode.Line:
				line_guide = new Line();
				line_guide.Stroke = new SolidColorBrush(currentColor);
				line_guide.X1 = line_guide.X2 = pt1.x;
				line_guide.Y1 = line_guide.Y2 = pt1.y;
				drawarea.Children.Add(line_guide);
				break;
			case DrawMode.Rect:
				rect_guide = new Rectangle();
				rect_guide.Fill = new SolidColorBrush(currentColor);
				rect_guide.Margin = new Thickness(pt1.x, pt1.y, 0, 0);
				rect_guide.Width = 1;
				rect_guide.Height = 1;
				drawarea.Children.Add(rect_guide);
				break;
			case DrawMode.Text:
				s1d13781.DrawText(txtSample.Text, pt1.x, pt1.y, currentColor);

				text_guide = new TextBlock();
				text_guide.Margin = new Thickness(pt1.x, pt1.y, 0, 0);
				text_guide.Text = txtSample.Text;
				text_guide.Foreground = new SolidColorBrush(currentColor);
				drawarea.Children.Add(text_guide);

				tmr = new DispatcherTimer();
				tmr.Interval = new TimeSpan(0, 0, 0, 750);
				tmr.Tick += (s, o) => {
					tmr.Stop();
					drawarea.Children.Remove(text_guide);
				};
				tmr.Start();
				break;
			case DrawMode.Bitmap:
				s1d13781.DrawBitmap(sampleImage, pt1.x, pt1.y, siWidth, siHeight);
				break;
			}

			pressed = true;
		}

		private void SetPosition(uPoint pt)
		{
			lblPos.Text = string.Format("({0:D},{1:D})", pt.x, pt.y);
		}

		private void SetPosition(uPoint pt1, uPoint pt2)
		{
			lblPos.Text = string.Format("({0:D},{1:D}) - ({2:D},{3:D})", pt1.x, pt1.y, pt2.x, pt2.y);
		}

		private void OnDrawAreaMoved(object sender, PointerRoutedEventArgs e)
		{
			if (pressed == false && drawmode != DrawMode.Bitmap) return;
			var pt = new uPoint(e.GetCurrentPoint(drawarea).Position);

			switch (drawmode) {
			case DrawMode.Pixel:
			case DrawMode.Text:
				SetPosition(pt);
				break;
			case DrawMode.Line:
				if(line_guide != null) {
					line_guide.X2 = pt.x;
					line_guide.Y2 = pt.y;
					SetPosition(pt1, pt);
				}else {
					SetPosition(pt);
				}
				break;
			case DrawMode.Rect:
				if(rect_guide != null) {
					int mx = (pt.x < pt1.x) ? pt.x : pt1.x;
					int my = (pt.y < pt1.y) ? pt.y : pt1.y;
					rect_guide.Width = Math.Abs(pt.x - pt1.x);
					rect_guide.Height = Math.Abs(pt.y - pt1.y);
					rect_guide.Margin = new Thickness(mx, my, 0, 0);

					SetPosition(pt1, pt);
				} else {
					SetPosition(pt);
				}
				break;
			case DrawMode.Bitmap:
				if(sampleImage != null) {
					if(rect_guide == null) {
						rect_guide = new Rectangle();
						rect_guide.Stroke = new SolidColorBrush(Colors.White);
						rect_guide.Width = siWidth;
						rect_guide.Height = siHeight;
						drawarea.Children.Add(rect_guide);
					}
					rect_guide.Margin = new Thickness(pt.x, pt.y, 0, 0);
				}
				break;
			default:
				SetPosition(pt);
				break;
			}
		}

		private void OnDrawAreaReleased(object sender, PointerRoutedEventArgs e)
		{
			if (pressed == false) return;

			var pt = new uPoint(e.GetCurrentPoint(drawarea).Position);
			SetPosition(pt);

			switch (drawmode) {
			case DrawMode.Line:
				s1d13781.DrawLine(pt1.x, pt1.y, pt.x, pt.y, currentColor);
				break;
			case DrawMode.Rect:
				if (pt1.x > pt.x) Swap(ref pt1.x, ref pt.x);
				if (pt1.y > pt.y) Swap(ref pt1.y, ref pt.y);
				s1d13781.FillRect(pt1.x, pt1.y, 
					(ushort)(pt.x - pt1.x), (ushort)(pt.y - pt1.y), currentColor);
				break;
			}

			ClearGuides();

			pressed = false;
		}

		private void Swap(ref ushort a, ref ushort b)
		{
			ushort c = a; a = b; b = c;
		}

		private void ClearGuides()
		{
			drawarea.Children.Clear();
			line_guide = null;
			rect_guide = null;
			text_guide = null;
		}
	}
}
