﻿using Microsoft.Graphics.Canvas;
using Microsoft.Graphics.Canvas.Text;
using Microsoft.IoT.Lightning.Providers;
using System;
using System.Diagnostics;
using System.Threading.Tasks;
using Windows.Devices;
using Windows.Devices.Enumeration;
using Windows.Devices.Gpio;
using Windows.Devices.Spi;
using Windows.Foundation;
using Windows.UI;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;

namespace TNKLabo
{
	namespace S1D13781
	{
		public sealed class Delay
		{
			private static Stopwatch sw = new Stopwatch();
			private static long mns;

			static Delay()
			{
				mns = (long)(1000000000.0 / Stopwatch.Frequency + 0.5);
			}

			public static long MinimalNanoSeconds
			{
				get { return mns; }
			}

			public static void DelayMicro(long microSeconds)
			{
				WaitLoop(microSeconds, false);
			}

			public static void DelayNano(long nanoSeconds)
			{
				WaitLoop(nanoSeconds, true);
			}

			private static void WaitLoop(long s, bool nano)
			{
				if (nano == false) s *= 1000;
				if (s < mns) s = mns;

				long ticks = s * Stopwatch.Frequency / 1000000000L;
				sw.Restart();
				while (sw.ElapsedTicks < ticks) ;
			}
		}

		public enum ImageDataFormat
		{
			RGB_888 = 0,  // RGB 8:8:8
			RGB_565 = 1,  // RGB 5:6:5
			Reserved1 = 2,
			Reserved2 = 3,
			RGB_888LUT = 4,
			RGB_565LUT = 5,
			RGB_332LUT = 6,  // RGB 3:3:2 Color
			Invalid = 7  // Invalid Format
		}

		[Flags]
		public enum DisplaySetting
		{
			DEFAULT = 0x00,

			// bit 0
			PANEL_DISABLE = 0x00,
			PANEL_ENABLE = 0x01,

			// bit 4
			DISPLAY_ON = 0x00,
			DISPLAY_OFF = 0x08,

			// bit 5,6
			TE_DISABLE = 0x00,
			TE_VNDP = 0x10,
			TE_LINECOUNT = 0x20,
			TE_BITBLT_BUSY_STATE = 0x30,

			// bit8
			TE_OUTPUT_ENABLE = 0x00,
			TE_OUTPUT_DISABLE = 0x80,
		}

		public struct iRect
		{
			public iRect(int x, int y, int w, int h)
			{
				this.x = x; this.y = y; this.w = w; this.h = h;
			}
			public int x, y, w, h;
			public int r { get { return x + w; } }
			public int b { get { return y + h; } }
		}

		public sealed class S1D13781
		{
			private enum RegIndex
			{
				REG00_REV_CODE = 0x00,     // Revision Code Register [READONLY]
				REG02_PROD_CODE = 0x02,     // Product Code Register [READONLY]
				REG04_POWER_SAVE = 0x04,     // Power Save Register
				REG06_RESET = 0x06,     // Software Reset Register
				REG10_PLL_0 = 0x10,     // PLL Setting Register 0
				REG12_PLL_1 = 0x12,     // PLL Setting Register 1
				REG14_PLL_2 = 0x14,     // PLL Setting Register 2
				REG16_INTCLK = 0x16,     // Internal Clock Configuration Register
				REG20_PANEL_SET = 0x20,     // Panel Setting Register
				REG22_DISP_SET = 0x22,     // Display Setting Register
				REG24_HDISP = 0x24,     // Horizontal Display Width Register
				REG26_HNDP = 0x26,     // Horizontal Non-Display Period Register
				REG28_VDISP = 0x28,     // Vertical Display Height Register
				REG2A_VNDP = 0x2A,     // Vertical Non-Display Period Register
				REG2C_HSW = 0x2C,     // HS Pulse Width Register
				REG2E_HPS = 0x2E,     // HS Pulse Start Position Register
				REG30_VSW = 0x30,     // VS Pulse Width Register
				REG32_VPS = 0x32,     // VS Pulse Start Position Register
				REG34_LINE_COUNT = 0x34,     // TE Line Count Register
				REG40_MAIN_SET = 0x40,     // Main Layer Setting Register
				REG42_MAIN_SADDR_0 = 0x42,     // Main Layer Start Address Register 0
				REG44_MAIN_SADDR_1 = 0x44,     // Main Layer Start Address Register 1
				REG46_MAIN_WIDTH = 0x46,     // Main Layer Width Register [READONLY]
				REG48_MAIN_HEIGHT = 0x48,     // Main Layer Height Register [READONLY]
				REG50_PIP_SET = 0x50,     // PIP Layer Setting Register
				REG52_PIP_SADDR_0 = 0x52,     // PIP Layer Start Address Register 0
				REG54_PIP_SADDR_1 = 0x54,     // PIP Layer Start Address Register 1
				REG56_PIP_WIDTH = 0x56,     // PIP Layer Width Register
				REG58_PIP_HEIGHT = 0x58,     // PIP Layer Height Register
				REG5A_PIP_XSTART = 0x5A,     // PIP Layer X Start Position Register
				REG5C_PIP_YSTART = 0x5C,     // PIP Layer Y Start Position Register
				REG60_PIP_EN = 0x60,     // PIP Enable Register
				REG62_ALPHA = 0x62,     // Alpha Blending Register
				REG64_TRANS = 0x64,     // Transparency Register
				REG66_KEY_0 = 0x66,     // Key Color Register 0
				REG68_KEY_1 = 0x68,     // Key Color Register 1
				REG80_BLT_CTRL_0 = 0x80,     // BitBLT Control Register 0 [WRITEONLY]
				REG82_BLT_CTRL_1 = 0x82,     // BitBLT Control Register 1
				REG84_BLT_STATUS = 0x84,     // BitBLT Status Register [READONLY]
				REG86_BLT_CMD = 0x86,     // BitBLT Command Register
				REG88_BLT_SSADDR_0 = 0x88,     // BitBLT Source Start Address Register 0
				REG8A_BLT_SSADDR_1 = 0x8A,     // BitBLT Source Start Address Register 1
				REG8C_BLT_DSADDR_0 = 0x8C,     // BitBLT Destination Start Address Register 0
				REG8E_BLT_DSADDR_1 = 0x8E,     // BitBLT Destination Start Address Register 1
				REG90_BLT_RECTOFFSET = 0x90,     // BitBLT Rectangle Offset Register
				REG92_BLT_WIDTH = 0x92,     // BitBLT Width Register
				REG94_BLT_HEIGHT = 0x94,     // BitBLT Height Register
				REG96_BLT_BGCOLOR_0 = 0x96,     // BitBLT Background Color Register 0
				REG98_BLT_BGCOLOR_1 = 0x98,     // BitBLT Background Color Register 1
				REG9A_BLT_FGCOLOR_0 = 0x9A,     // BitBLT Foreground Color Register 0
				REG9C_BLT_FGCOLOR_1 = 0x9C,     // BitBLT Foreground Color Register 1
				REGD0_GPIO_CONFIG = 0xD0,     // GPIO Configuration Register
				REGD2_GPIO_STATUS = 0xD2,     // GPIO Status / Control Register
				REGD4_GPIO_PULLDOWN = 0xD4,     // GPIO Pull-Down Control Register

				REGFLAG_BASE = 0xFC,     // Special reserved flags beyond this point
				REGFLAG_DELAY = 0xFC,     // PLL Register Programming Delay (in us)
				S1D_REGDELAYPLL = 0xFC,     // PLL Register Programming Delay (in us)  added for 781cfg
				REGFLAG_OFF_DELAY = 0xFD,     // LCD Panel Power Off Delay (in ms)
				S1D_REGDELAYOFF = 0xFD,     // LCD Panel Power Off Delay (in ms)  added for 781cfg
				REGFLAG_ON_DELAY = 0xFE,     // LCD Panel Power On Delay (in ms)
				S1D_REGDELAYON = 0xFE,     // LCD Panel Power On Delay (in ms)  added for 781cfg
				REGFLAG_END_OF_TABLE = 0xFF,     // End of Registers Marker
			}

			private struct RegData
			{
				public RegIndex regIndex;  //Register index
				public ushort regValue;  //Register data
				public RegData(RegIndex ri, ushort rv)
				{
					regIndex = ri;
					regValue = rv;
				}
			}

			private RegData[] regInitValues = new RegData[] {
				new RegData(RegIndex.REG06_RESET, 0x0100),
				new RegData(RegIndex.S1D_REGDELAYON, 0x2710),
				new RegData(RegIndex.REG04_POWER_SAVE, 0x0000),
				new RegData(RegIndex.REG10_PLL_0, 0x0000),
				new RegData(RegIndex.REG12_PLL_1, 0x0000),
				new RegData(RegIndex.REG14_PLL_2, 0x003E),
				new RegData(RegIndex.REG10_PLL_0, 0x0001),
				new RegData(RegIndex.S1D_REGDELAYPLL, 0x09C4),
				new RegData(RegIndex.REG16_INTCLK, 0x0006),
				new RegData(RegIndex.REG04_POWER_SAVE, 0x0001),
				new RegData(RegIndex.S1D_REGDELAYPLL, 0x09C4),
				new RegData(RegIndex.REG20_PANEL_SET, 0x004F),
				new RegData(RegIndex.REG22_DISP_SET, 0x0001),
				new RegData(RegIndex.REG24_HDISP, 0x003C),
				new RegData(RegIndex.REG26_HNDP, 0x002D),
				new RegData(RegIndex.REG28_VDISP, 0x0110),
				new RegData(RegIndex.REG2A_VNDP, 0x0010),
				new RegData(RegIndex.REG2C_HSW, 0x0001),
				new RegData(RegIndex.REG2E_HPS, 0x0005),
				new RegData(RegIndex.REG30_VSW, 0x0001),
				new RegData(RegIndex.REG32_VPS, 0x0008),
				new RegData(RegIndex.REG40_MAIN_SET, 0x0000),
				new RegData(RegIndex.REG42_MAIN_SADDR_0, 0x0000),
				new RegData(RegIndex.REG44_MAIN_SADDR_1, 0x0000),
				new RegData(RegIndex.REG50_PIP_SET, 0x0001),
				new RegData(RegIndex.REG52_PIP_SADDR_0, 0xFA00),
				new RegData(RegIndex.REG54_PIP_SADDR_1, 0x0005),
				new RegData(RegIndex.REG56_PIP_WIDTH, 0x0020),
				new RegData(RegIndex.REG58_PIP_HEIGHT, 0x0018),
				new RegData(RegIndex.REG5A_PIP_XSTART, 0x00A0),
				new RegData(RegIndex.REG5C_PIP_YSTART, 0x0000),
				new RegData(RegIndex.REG60_PIP_EN, 0x0000),
				new RegData(RegIndex.REG62_ALPHA, 0x0040),
				new RegData(RegIndex.REG64_TRANS, 0x0000),
				new RegData(RegIndex.REG66_KEY_0, 0x0000),
				new RegData(RegIndex.REG68_KEY_1, 0x0000),
				new RegData(RegIndex.REGD0_GPIO_CONFIG, 0x0000),
				new RegData(RegIndex.REGD2_GPIO_STATUS, 0x0000),
				new RegData(RegIndex.REGD4_GPIO_PULLDOWN, 0x000E),
				new RegData(RegIndex.REG04_POWER_SAVE, 0x0002),
				new RegData(RegIndex.REGFLAG_END_OF_TABLE, 0x0000)
			};

			private SpiDevice SPI = null;
			private GpioPin CsPin = null;

			private enum SPIBIT
			{
				SPIWRITE_8BIT = 0x80,
				SPIREAD_8BIT = 0xC0,
				SPIWRITE_16BIT = 0x88,
				SPIREAD_16BIT = 0xC8
			}

			private void RegWrite(RegIndex regIndex, ushort regValue)
			{
				if (CsPin == null) return;

				CsPin.Write(GpioPinValue.Low);

				SPI.Write(new byte[]{
					(byte)SPIBIT.SPIWRITE_8BIT, //command
					0x06, //address bits 18:16
					0x08, //address bits 15:8
					(byte)regIndex, // address bits 7:0
					(byte)(regValue & 0x00FF), //data bits 7:0
					(byte)(regValue >> 8) //data bits 15:8
				});

				CsPin.Write(GpioPinValue.High);
			}

			private ushort RegRead(RegIndex regIndex)
			{
				if (CsPin == null) return 0x00;

				CsPin.Write(GpioPinValue.Low);

				SPI.Write(new byte[]{
					(byte)SPIBIT.SPIREAD_8BIT, //command
					0x06, //address bits 18:16
					0x08, //address bits 15:8
					(byte)regIndex, // address bits 7:0
				});

				byte[] res = new byte[3]; // 0:dummy, 1-2:data
				SPI.Read(res);

				CsPin.Write(GpioPinValue.High);

				return (ushort)(res[1] | (res[2] << 8));
			}

			private uint vramAddress = 0;
			private ImageDataFormat imageFormat = ImageDataFormat.Invalid;
			private ushort stride = 0, bytesPerPixel = 0, width = 0, height = 0;

			public ushort Width { get { return width; } }
			public ushort Height { get { return height; } }
			public ImageDataFormat ImageFormat { get { return imageFormat; } }

			private void InitRegs()
			{
				for (int i = 0; regInitValues[i].regIndex != RegIndex.REGFLAG_END_OF_TABLE; i++) {
					RegIndex index = regInitValues[i].regIndex;

					if (index < RegIndex.REGFLAG_BASE) {
						RegWrite(index, regInitValues[i].regValue);
					} else if (index == RegIndex.S1D_REGDELAYOFF || index == RegIndex.S1D_REGDELAYON) {
						Delay.DelayMicro(regInitValues[i].regValue);
					} else if (index == RegIndex.S1D_REGDELAYPLL) {
						if ((RegRead(RegIndex.REG10_PLL_0) & 0x0002) == 0) { // don't poll PLL Lock status bit if PLL Bypass is enabled
							while ((RegRead(RegIndex.REG10_PLL_0) & 0x8000) == 0) {
								Task.Delay(100).Wait();
							}
						}
					}
				}

				width = RegRead(RegIndex.REG46_MAIN_WIDTH);
				height = RegRead(RegIndex.REG48_MAIN_HEIGHT);

				switch (RegRead(RegIndex.REG40_MAIN_SET) & 0x0007) {
					case 0: imageFormat = ImageDataFormat.RGB_888; break;
					case 1: imageFormat = ImageDataFormat.RGB_565; break;
					case 4: imageFormat = ImageDataFormat.RGB_888LUT; break;
					case 5: imageFormat = ImageDataFormat.RGB_565LUT; break;
					case 6: imageFormat = ImageDataFormat.RGB_332LUT; break;
					default: imageFormat = ImageDataFormat.RGB_888; break;
				}
				switch (imageFormat) {
				case ImageDataFormat.RGB_332LUT:
					bytesPerPixel = 0x01;
					break;
				case ImageDataFormat.RGB_565:
				case ImageDataFormat.RGB_565LUT:
					bytesPerPixel = 0x02;
					break;
				case ImageDataFormat.RGB_888:
				case ImageDataFormat.RGB_888LUT:
					bytesPerPixel = 0x03;
					break;
				default:
					bytesPerPixel = 0x00;
					break;
				}
				stride = (ushort)(width * bytesPerPixel);

				vramAddress = RegRead(RegIndex.REG42_MAIN_SADDR_0) | (uint)(RegRead(RegIndex.REG44_MAIN_SADDR_1) << 16);

				Clear();
			}

			private void SetDisplayOption(DisplaySetting setting)
			{
				RegWrite(RegIndex.REG22_DISP_SET, (ushort)setting);
			}

			public int CsPinNo { get; set; }

			public S1D13781()
			{
				CsPinNo = 24;
			}

			public async void Begin()
			{
				var settings = new SpiConnectionSettings(0);
				settings.ClockFrequency = 24000000;
				settings.Mode = SpiMode.Mode0;
				settings.DataBitLength = 8;

				if (LightningProvider.IsLightningEnabled) {
					LowLevelDevicesController.DefaultProvider = LightningProvider.GetAggregateProvider();

					var cs = await SpiController.GetControllersAsync(LightningSpiProvider.GetSpiProvider());
					var controller = cs[0];
					SPI = controller.GetDevice(settings);

					var gpc = await GpioController.GetControllersAsync(LightningGpioProvider.GetGpioProvider());
					CsPin = gpc[0].OpenPin(CsPinNo);
				} else {
					string aqs = SpiDevice.GetDeviceSelector("SPI0");
					var dis = DeviceInformation.FindAllAsync(aqs).GetResults();
					if(dis.Count > 0) {
						SPI = await SpiDevice.FromIdAsync(dis[0].Id, settings);
						CsPin = GpioController.GetDefault().OpenPin(CsPinNo);
					}
				}
				if (CsPin != null) {
					CsPin.SetDriveMode(GpioPinDriveMode.Output);
					InitRegs();
				}
			}

			private void MemWriteByte(uint memAddress, byte memValue)
			{
				CsPin.Write(GpioPinValue.Low);

				SPI.Write(new byte[] {
					(byte)SPIBIT.SPIWRITE_8BIT, //command
					(byte)(memAddress >> 16), //address bits 18:16
					(byte)((memAddress & 0x0000FF00) >> 8), //address bits 15:8
					(byte)(memAddress & 0x000000FF), //address bits 7:0
					memValue, //data bits 7:0
				});

				CsPin.Write(GpioPinValue.High);
			}

			void MemWriteWord(uint memAddress, ushort memValue)
			{
				CsPin.Write(GpioPinValue.Low);

				SPI.Write(new byte[] {
					(byte)SPIBIT.SPIWRITE_8BIT, //command
					(byte)(memAddress >> 16), //address bits 18:16
					(byte)((memAddress & 0x0000FF00) >> 8), //address bits 15:8
					(byte)(memAddress & 0x000000FF), //address bits 7:0
					(byte)(memValue & 0x00FF), //data bits 7:0
					(byte)(memValue >> 8) //data bits 15:8
				});

				CsPin.Write(GpioPinValue.High);
			}

			private bool PollWhileRegBitsHigh(RegIndex regIndex, ushort regBits, uint intervalMs, uint timeoutMs)
			{
				bool result = false;

				if (intervalMs == 0) {
					Stopwatch _sw = new Stopwatch();
					_sw.Start();
					int startTimeMs = _sw.Elapsed.Milliseconds;  //get the start time in milliseconds
					do {
						if ((RegRead(regIndex) & regBits) == 0) {
							result = true;
							break;
						}
					} while ((_sw.Elapsed.Milliseconds - startTimeMs) < timeoutMs);
				} else {
					uint timer = 0;
					do {
						if ((RegRead(regIndex) & regBits) == 0) {
							result = true;
							break;
						}
						Delay.DelayMicro(intervalMs);
						timer += intervalMs;
					} while (timer < timeoutMs);
				}

				return result;
			}

			private uint ConvertColor(Color color)
			{
				uint c = 0x00000000;
				switch (imageFormat) {
				case ImageDataFormat.RGB_888:
				case ImageDataFormat.RGB_888LUT:
					c = (uint)((color.R << 16) | (color.G << 8) | color.B);
					break;

				case ImageDataFormat.RGB_565:
				case ImageDataFormat.RGB_565LUT:
					c = (uint)(((color.R >> 3) << 11) | ((color.G >> 2) << 5) | (color.B >> 3));
					break;

				case ImageDataFormat.RGB_332LUT:
					c = (uint)((color.R * 7 / 255) << 5 + (color.G * 7 / 255) << 2 + (color.B * 3 / 255));
					break;

				default:
					throw new NotSupportedException();
				}
				return c;
			}

			public void Clear()
			{
				FillRect(0, 0, width, height, Colors.Black);
			}

			public void FillRect(ushort xStart, ushort yStart, ushort width, ushort height, Color color)
			{
				if (CsPin == null) return;

				uint c = ConvertColor(color);
				byte dataType = 0;
				switch (imageFormat) {
				case ImageDataFormat.RGB_888:
					dataType = 2;
					break;
				case ImageDataFormat.RGB_565:
					dataType = 1;
					break;
				default:
					throw new NotSupportedException();
				}

				//set correct memory address for BitBLT
				uint vadd = (uint)(vramAddress + (yStart * stride) + (xStart * bytesPerPixel));

				//setup the BitBLT registers
				RegWrite(RegIndex.REG82_BLT_CTRL_1, (ushort)(dataType << 2));
				RegWrite(RegIndex.REG86_BLT_CMD, 0x0002);  //Solid Fill BitBLT
				RegWrite(RegIndex.REG8C_BLT_DSADDR_0, (ushort)(vadd & 0xFFFF));
				RegWrite(RegIndex.REG8E_BLT_DSADDR_1, (ushort)((vadd >> 16) & 0x0007));
				RegWrite(RegIndex.REG90_BLT_RECTOFFSET, (ushort)(stride / bytesPerPixel)); // 781 requires stride in pixels (not bytes)
				RegWrite(RegIndex.REG92_BLT_WIDTH, width);
				RegWrite(RegIndex.REG94_BLT_HEIGHT, height);

				RegWrite(RegIndex.REG9A_BLT_FGCOLOR_0, (ushort)(c & 0xFFFF));
				RegWrite(RegIndex.REG9C_BLT_FGCOLOR_1, (ushort)((c >> 16) & 0x00FF));

				// Start the BitBLT
				RegWrite(RegIndex.REG80_BLT_CTRL_0, 0x0001);

				// Wait for completion.
				PollWhileRegBitsHigh(RegIndex.REG84_BLT_STATUS, 0x0001, 0, 1000);
			}

			public void DrawPixel(int x, int y, Color color)
			{
				if (x < 0 || y < 0 || x >= width || y >= height) return;

				uint offset = (uint)(vramAddress + (y * stride) + (x * bytesPerPixel));

				switch (imageFormat) {
				case ImageDataFormat.RGB_888:
				case ImageDataFormat.RGB_888LUT:
					//write the color components
					MemWriteByte(offset, color.B);
					MemWriteByte(offset + 1, color.G);
					MemWriteByte(offset + 2, color.R);
					break;

				case ImageDataFormat.RGB_565:
				case ImageDataFormat.RGB_565LUT:
					MemWriteWord(offset, (ushort)(((color.R >> 3) << 11) | ((color.G >> 2) << 5) | (color.B >> 3)));
					break;

				case ImageDataFormat.RGB_332LUT:
					MemWriteByte(offset, (byte)((color.R * 7 / 255) << 5 + (color.G * 7 / 255) << 2 + (color.B * 3 / 255)));
					break;

				default:
					throw new NotSupportedException();
				}
			}

			private void Swap(ref int a, ref int b)
			{
				int c = a;
				a = b;
				b = c;
			}

			public void DrawLine(ushort x1, ushort y1, ushort x2, ushort y2, Color color)
			{
				if (x1 >= width && x2 >= width) return;
				if (y1 >= height && y2 >= height) return;

				const int fixfloat_bit = 20;

				if(Math.Abs(x1 - x2) > Math.Abs(y1 - y2)) {
					int fy = y1 << fixfloat_bit;
					int dy = 0;
					if (x1 != x2) dy = ((y2 - y1) << fixfloat_bit) / (x2 - x1);

					int a = 1;
					if(x1 > x2) { a = -1; dy = -dy; }

					while(x1 != x2) {
						DrawPixel(x1, fy >> fixfloat_bit, color);
						fy += dy;
						x1 = (ushort)(x1 + a);
						if (x1 >= width) break;
					}
				}else {
					int fx = x1 << fixfloat_bit;
					int dx = 0;
					if (y1 != y2) dx = ((x2 - x1) << fixfloat_bit) / (y2 - y1);

					int a = 1;
					if (y1 > y2) { a = -1; dx = -dx; }

					while (y1 != y2) {
						DrawPixel(fx >> fixfloat_bit, y1, color);
						fx += dx;
						y1 = (ushort)(y1 + a);
						if (y1 >= height) break;
					}

				}
			}

			private bool IsRectOverlap(iRect src, iRect dest)
			{
				return (src.x < dest.r && src.r > dest.x && src.y < dest.b && src.b > dest.y);
			}

			public void BitBlt(iRect src, int dx, int dy)
			{
				ushort datatype = 0;
				switch (imageFormat) {
				case ImageDataFormat.RGB_888:
				case ImageDataFormat.RGB_888LUT:
					datatype = 2;
					break;
				case ImageDataFormat.RGB_565:
				case ImageDataFormat.RGB_565LUT:
					datatype = 1;
					break;
				default:
					return;
				}

				if (src.w >= width) src.w = width;
				if (src.r >= width) src.x = width - src.w;
				if (src.x < 0) src.x = 0;

				if (src.h >= height) src.h = height;
				if (src.b >= height) src.y = height - src.h;
				if (src.y < 0) src.y = 0;

				ushort cmd, src_start, dest_start;
				iRect dest = new iRect(dx, dy, src.w, src.h);
				// Check to see if regions are overlapping in a way that requires a reverse copy.
				if (dy >= src.y && IsRectOverlap(src, dest)) {
					cmd = 1; // Move Negative
					src_start = (ushort)(vramAddress +(src.b - 1) * stride + src.r * bytesPerPixel);
					dest_start = (ushort)(vramAddress + (dest.y + src.h - 1) * stride + (dest.x + src.w) * bytesPerPixel);
				} else {
					cmd = 0; // Move Positive
					src_start = (ushort)(vramAddress + src.y * stride + src.x * bytesPerPixel);
					dest_start = (ushort)(vramAddress + dest.y * stride + dest.x * bytesPerPixel);
				}

				//setup the BitBLT parameters
				RegWrite(RegIndex.REG82_BLT_CTRL_1, (ushort)(datatype << 2));
				RegWrite(RegIndex.REG86_BLT_CMD, cmd);
				RegWrite(RegIndex.REG88_BLT_SSADDR_0, (ushort)(src_start & 0xFFFF));
				RegWrite(RegIndex.REG8A_BLT_SSADDR_1, (ushort)((src_start >> 16) & 0x0007));
				RegWrite(RegIndex.REG8C_BLT_DSADDR_0, (ushort)(dest_start & 0xFFFF));
				RegWrite(RegIndex.REG8E_BLT_DSADDR_1, (ushort)((dest_start >> 16) & 0x0007));
				RegWrite(RegIndex.REG90_BLT_RECTOFFSET, (ushort)(stride / bytesPerPixel)); // 781 requires stride in pixels (not bytes)
				RegWrite(RegIndex.REG92_BLT_WIDTH, (ushort)src.w);
				RegWrite(RegIndex.REG94_BLT_HEIGHT, (ushort)src.h);

				// Start the BitBLT
				RegWrite(RegIndex.REG80_BLT_CTRL_0, 0x0001);

				// Wait for completion.
				PollWhileRegBitsHigh(RegIndex.REG84_BLT_STATUS, 0x0001, 0, 1000);
			}

			private FontFamily font = FontFamily.XamlAutoFontFamily;
			private double fontsize = 12.0;
			public FontFamily Font
			{
				set { font = value; }
			}
			public double FontSize
			{
				set { fontsize = value; }
			}

			public Size MeasureText(string text)
			{
				TextBlock tb = new TextBlock();
				tb.Text = text;
				tb.FontFamily = font;
				tb.FontSize = fontsize;
				tb.Measure(new Size(double.PositiveInfinity, double.PositiveInfinity));
				return tb.DesiredSize;
			}

			public void DrawText(string text, ushort x, ushort y, Color color)
			{
				Size sz = MeasureText(text);
				int w = (int)sz.Width + 10, h = (int)sz.Height + 10;

				CanvasTextFormat ctf = new CanvasTextFormat();
				ctf.FontFamily = font.Source;
				ctf.FontSize = (float)fontsize;
				ctf.HorizontalAlignment = CanvasHorizontalAlignment.Left;
				ctf.VerticalAlignment = CanvasVerticalAlignment.Top;
				ctf.WordWrapping = CanvasWordWrapping.NoWrap;

				var device = CanvasDevice.GetSharedDevice();
				CanvasBitmap cb = CanvasBitmap.CreateFromBytes(device,
					new byte[w * h * 4], w, h, Windows.Graphics.DirectX.DirectXPixelFormat.B8G8R8A8UIntNormalized);

				var offscreen = new CanvasRenderTarget(device, w, h, 96);
				using (var ds = offscreen.CreateDrawingSession()) {
					ds.TextAntialiasing = Microsoft.Graphics.Canvas.Text.CanvasTextAntialiasing.Aliased;
					ds.DrawText(text, new Rect(0, 0, w, h), Colors.White, ctf);
				}

				byte[] buffer = offscreen.GetPixelBytes();
				for (int yy = 0; yy < h; yy++) {
					if (y + yy >= height) break;

					int p = yy * w * 4;
					for (int xx = 0; xx < w; xx++) {
						if (x + xx >= width) break;
						if (buffer[p] != 0x00) DrawPixel(x + xx, y + yy, color);
						p += 4;
					}
				}
			}

			public void DrawBitmap(byte[] bmpbuffer, int sx, int sy, int bw, int bh)
			{
				byte dataType = 0;
				switch (imageFormat) {
				case ImageDataFormat.RGB_888:
				case ImageDataFormat.RGB_888LUT:
					dataType = 2;
					break;
				case ImageDataFormat.RGB_565:
				case ImageDataFormat.RGB_565LUT:
					dataType = 1;
					break;
				default:
					throw new NotSupportedException();
				}

				int linelen = bw * 3 + 4;
				byte[] sendbuf = new byte[linelen];
				for (ushort y = 0; y < bh; y++) {
					if (sy + y >= height) break;

					uint va = (uint)(vramAddress + ((sy + y) * width + sx) * 3);
					sendbuf[0] = (byte)SPIBIT.SPIWRITE_8BIT; //command
					sendbuf[1] = (byte)(va >> 16); //address bits 18:16
					sendbuf[2] = (byte)((va & 0x0000FF00) >> 8); //address bits 15:8
					sendbuf[3] = (byte)(va & 0x000000FF); //address bits 7:0

					int sendbuf_p = 4, bgrabuf_p = y * bw * 4;

					for (ushort x = 0; x < bw; x++) {
						if (sx + x >= width) break;

						if (dataType == 2) {
							sendbuf[sendbuf_p++] = bmpbuffer[bgrabuf_p++];
							sendbuf[sendbuf_p++] = bmpbuffer[bgrabuf_p++];
							sendbuf[sendbuf_p++] = bmpbuffer[bgrabuf_p++];
							bgrabuf_p++;
						}
					}

					CsPin.Write(GpioPinValue.Low);
					SPI.Write(sendbuf);
					CsPin.Write(GpioPinValue.High);
				}
			}
		}
	}
}
